from .basis import Basis
from .path import Path
from .borehole import Borehole, SingleUTube, MultipleUTube
from .borefield import Borefield, Network
from .ground_heat_exchanger import UTubeHeatExchanger, CoaxialHeatExchanger
from .thermal_resistance import Multipole
from .simulation import LoadAggregation, Simulation, LoadHistoryReconstruction, gFunction
from . import thermal_resistance
from . import utilities
from . import plot



# -*- coding: utf-8 -*-
from functools import partial

from jax import numpy as jnp
from jax import Array, jacobian, jit, vmap
from jax.typing import ArrayLike
from jax.scipy.special import erfc
import numpy as np
from scipy.special import roots_legendre


class Path:
    """Trajectory of a borehole.

    Parameters
    ----------
    xi : array_like
        (`n_nodes`,) array of node coordinates along the interval
        ``[-1, 1]``.
    p : array_like
        Positions (``x``, ``y``, ``z``) of the nodes along the trajectory
        of the borehole.
    order : int or None, default: None
        Order of the polynomial regression through positions `p` to
        approximate the borehole trajectory. The first row of `p`, which
        should correspond to the top positition of the borehole (i.e.
        with ``xi[0]=-1``), is constrained in the polynomial regression.
        If `order` is ``None``, then ``order=len(xi)``.
    s_order : int or None, default: None
        Order of the polynomial approximant of the longitudinal position
        ``s`` along the borehole trajectory. If `s_order` is ``None``,
        then ``s_order=2*len(xi)-1``.

    Attributes
    ----------
    n_nodes : int
        Number of nodes.

    """

    def __init__(self, xi: ArrayLike, p: ArrayLike, order: int | None = None, s_order: int | None = None):
        # Runtime type validation
        if not isinstance(xi, ArrayLike):
            raise TypeError(f"Expected arraylike input; got {xi}")
        if not isinstance(p, ArrayLike):
            raise TypeError(f"Expected arraylike input; got {p}")
        # Convert input to jax.Array
        xi = jnp.asarray(xi)
        p = jnp.atleast_2d(p)

        # --- Class atributes ---
        self.xi = xi
        self.p = p
        n_nodes = len(xi)
        self.n_nodes = n_nodes
        if order is None:
            order = n_nodes
        order = jnp.maximum(2, jnp.minimum(order, n_nodes))
        self.order = order
        if s_order is None:
            s_order = 2 * (self.n_nodes - 1)
        self.s_order = s_order

        # --- Path functions ---
        # Position (p)
        p_w = np.ones(n_nodes)
        if order < n_nodes:
            p_w[0] = 1e4
        p_coefs = jnp.array(np.polyfit(xi, p, order-1, w=p_w))
        f_p = lambda _eta: jnp.polyval(p_coefs, _eta)
        self.f_p = jit(
            lambda _eta: vmap(f_p, in_axes=0)(_eta) if len(jnp.shape(_eta)) > 0 else f_p(_eta)
        )
        # Derivative of position (dp/dxi)
        f_dp_dxi = lambda _eta: jacobian(f_p)(_eta)
        self.f_dp_dxi = jit(
            lambda _eta: vmap(f_dp_dxi, in_axes=0)(_eta) if len(jnp.shape(_eta)) > 0 else f_dp_dxi(_eta)
        )
        # Norm of the Jacobian (J)
        f_J = lambda _eta: jnp.linalg.norm(f_dp_dxi(_eta))
        self.f_J = jit(
            lambda _eta: vmap(f_J, in_axes=0)(_eta) if len(jnp.shape(_eta)) > 0 else f_J(_eta)
        )
        # Longitudinal position (s)
        x_s, w_s = roots_legendre(s_order)
        x_s = jnp.array(x_s)
        w_s = jnp.array(w_s)
        low = jnp.concatenate([-jnp.ones(1), x_s[:-1]])
        high = x_s
        s = jnp.cumsum(
            vmap(
                lambda _a, _b: self.f_J(0.5 * (_a + _b) + 0.5 * x_s * (_b - _a)) @ w_s * 0.5 * (_b - _a),
                in_axes=(0, 0)
            )(low, high)
        )
        s_coefs = jnp.array(np.polyfit(x_s, s, s_order-1))
        f_s = lambda _eta: jnp.polyval(s_coefs, _eta)
        self.f_s = jit(
            lambda _eta: vmap(f_s, in_axes=0)(_eta) if len(jnp.shape(_eta)) > 0 else f_s(_eta)
        )

    @partial(jit, static_argnames=['self'])
    def point_heat_source(self, xi: Array | float, p: Array, time: Array | float, alpha: float, r_min: float = 0.) -> Array | float:
        """Point heat source solution.

        Parameters
        ----------
        xi : array or float
            (N,) array of the coordinates of the point heat sources along
            the trajectory.
        p : array
            (M, 3,) array of the positions at which the point heat source
            solution is evaluated.
        time : array or float
            (K,) array of times (in seconds).
        alpha : float
            Ground thermal diffusivity (in m^2/s).
        r_min : float, default: ``0.``
            Minimum distance (in meters) between point heat sources and
            positions `p`.

        Returns
        -------
        array or float
            (K, M, N,) array of values of the point heat source solution.
            For each of the parameters `xi`, `p` and `time`, the
            corresponding axis is removed if the parameter is supplied as
            a ``float``.

        """
        if len(jnp.shape(time)) > 0:
            return vmap(
                self.point_heat_source,
                in_axes=(None, None, -1, None, None)
            )(xi, p, time, alpha, r_min)
        if len(jnp.shape(p)) > 1:
            return vmap(
                self.point_heat_source,
                in_axes=(None, -2, None, None, None)
            )(xi, p, time, alpha, r_min)
        if len(jnp.shape(xi)) > 0:
            return vmap(
                self.point_heat_source,
                in_axes=(-1, None, None, None, None)
            )(xi, p, time, alpha, r_min)
        # Current position of the point source
        p_source = self.f_p(xi)
        # Distance to the real point (p)
        r = jnp.sqrt(jnp.linalg.norm(p_source - p)**2 + r_min**2)
        # Distance to the mirror point (p')
        r_mirror = jnp.linalg.norm(p_source - p * jnp.array([1, 1, -1]))
        # Point heat source solution
        h = 0.5 * erfc(r / jnp.sqrt(4 * alpha * time)) / r - 0.5 * erfc(r_mirror / jnp.sqrt(4 * alpha * time)) / r_mirror
        return h * self.f_J(xi)



# -*- coding: utf-8 -*-
from collections.abc import Callable 
from typing import Self, Tuple

from jax import numpy as jnp
from jax import Array, jit, vmap
from jax.typing import ArrayLike
import numpy as np
from scipy.special import roots_legendre


class Basis:
    """Polynomial basis functions.

    Parameters
    ----------
    xi : array_like
        (`n_nodes`,) array of node coordinates along the interval
        ``[-1, 1]``.
    order : int
        Order of the numerical quadratures over any interval ``[a, b]``.
    order_nodes : int
        Order of the numerical quadratures over the full interval
        ``[-1, 1]`` that use the node coordinates `xi` as stop points.
    w : array_like or None, default: None
        Quadrature weights for numerical integration using `xi` as
        quadrature points. If `w` is ``None``, the weights are evaluated
        through numerical integration of order `order`.

    Attributes
    ----------
    n_nodes : int
        Number of nodes.

    """

    def __init__(self, xi: ArrayLike, order: int = 101, order_nodes: int = 21, w: ArrayLike | None = None):
        # Runtime type validation
        if not isinstance(xi, ArrayLike):
            raise TypeError(f"Expected arraylike input; got {xi}")
        # Convert input to jax.Array
        xi = jnp.asarray(xi)

        # --- Class atributes ---
        self.xi = xi
        self.n_nodes = len(xi)
        self.order = order
        self.order_nodes = order_nodes

        # --- Basis functions (psi) ---
        power = jnp.arange(self.n_nodes)
        coefs = jnp.linalg.inv(
            np.power.outer(self.xi, power)
        )
        f_psi = lambda _eta: _eta**power @ coefs
        self.f_psi = jit(
            lambda _eta: vmap(f_psi, in_axes=0)(_eta) if len(jnp.shape(_eta)) > 0 else f_psi(_eta)
        )
        self.f = jit(
            lambda _eta, f_nodes: self.f_psi(_eta) @ f_nodes
        )

        # --- Initialize quadrature methods ---
        self._initialize_quad_methods(order, order_nodes)
        # Integration weights
        if w is None:
            w = self.f_psi(self._x_gl).T @ self._w_gl
        self.w = w

    def _initialize_quad_methods(self, order: int, order_nodes: int):
        """Initialize the points and weights for quadrature methods.

        Parameters
        ----------
        order : int
            Order of the numerical quadratures over any interval
            ``[a, b]``.
        order_nodes : int
            Order of the numerical quadratures over the full interval
            ``[-1, 1]`` that use the node coordinates `xi` as stop points.

        """
        self._x_gl, self._w_gl = self._gauss_legendre_rule(order)
        self._x_gl_nodes, self._w_gl_nodes, self._psi_gl_nodes = self._initialize_quad_nodes(*self._gauss_legendre_rule(order_nodes))
        self._x_ts, self._w_ts = self._tanh_sinh_rule(order)
        self._x_ts_nodes, self._w_ts_nodes, self._psi_ts_nodes = self._initialize_quad_nodes(*self._tanh_sinh_rule(order_nodes))

    def _initialize_quad_nodes(self, x: ArrayLike, w: ArrayLike) -> Tuple[Array, Array, Array]:
        """Points and weights for quadrature with stop points at nodes.

        Parameters
        ----------
        x : array_like
            Quadrature points over an interval ``[-1, 1]``.
        w : array_like
            Quadrature weights associated with `x`.

        Returns
        -------
        x_nodes : array
            Quadrature points for integration with stop points at nodes.
        w_nodes : array
            Quadrature weights for integration with stop points at nodes.
        psi_nodes : array
            Values of the basis functions at quadrature points.

        """
        interval = jnp.concatenate([-jnp.ones(1), self.xi, jnp.ones(1)])
        low, high = interval[:-1], interval[1:]
        x_nodes = jnp.concatenate([0.5 * (b + a) + 0.5 * (b - a) * x for a, b in zip(low, high)])
        w_nodes = jnp.concatenate([0.5 * (b - a) * w for a, b in zip(low, high)])
        psi_nodes = self.f_psi(x_nodes)
        return x_nodes, w_nodes, psi_nodes

    def _quad(self, fun: Callable[[Array], Array], a: float, b: float, x: Array, w: Array) -> Array:
        """Integral of a function multiplied with basis functions.

        Parameters
        ----------
        fun : callable
            Function to integrate over the interval ``[a, b]``.
        a, b : float
            Lower (`a`) and upper (`b`) bounds of integration.
        x : array
            Quadrature points.
        w : array
            Quadrature weights.

        Returns
        -------
        array
            Integral of `fun` multiplied by the basis functions over the
            interval ``[a, b]``.

        """
        x = 0.5 * (b + a) + 0.5 * (b - a) * x
        w = 0.5 * (b - a) * w
        psi = self.f_psi(x)
        return self._quad_psi(fun, x, w, psi)

    def _quad_psi(self, fun: Callable[[Array | float], Array | float], x: Array, w: Array, psi: Array) -> Array:
        """Multiply function with basis and evaluate quadrature.

        Parameters
        ----------
        fun : callable
            Function to integrate.
        x : array
            Quadrature points.
        w : array
            Quadrature weights.
        psi : array
            Basis functions evaluated at quadrature points.

        Returns
        -------
        array
            Integral of `fun` multiplied by the basis functions.

        """
        integral = lambda _eta, _w, _psi: (fun(_eta) * _psi) @ _w
        return vmap(integral, in_axes=(None, None, 1), out_axes=-1)(x, w, psi)

    def quad_gl(self, fun: Callable[[Array | float], Array | float], a: float, b: float) -> Array:
        """Gauss-Legendre quadrature.

        Parameters
        ----------
        fun : callable
            Function to integrate over the interval ``[a, b]``.
        a, b : float
            Lower (`a`) and upper (`b`) bounds of integration.

        Returns
        -------
        array
            Integral of `fun` multiplied by the basis functions over the
            interval ``[a, b]``.

        """
        x, w = self._x_gl, self._w_gl
        return self._quad(fun, a, b, x, w)

    def quad_gl_nodes(self, fun: Callable[[Array | float], Array | float]) -> Array:
        """Gauss-Legendre quadrature with stop points at nodes.

        Parameters
        ----------
        fun : callable
            Function to integrate over the interval ``[-1, 1]``.

        Returns
        -------
        array
            Integral of `fun` multiplied by the basis functions over the
            interval ``[-1, 1]``.

        """
        x, w, psi = self._x_gl_nodes, self._w_gl_nodes, self._psi_gl_nodes
        return self._quad_psi(fun, x, w, psi)

    def quad_ts_nodes(self, fun: Callable[[Array | float], Array | float]) -> Array:
        """tanh-sinh quadrature with stop points at nodes.

        Parameters
        ----------
        fun : callable
            Function to integrate over the interval ``[-1, 1]``.

        Returns
        -------
        array
            Integral of `fun` multiplied by the basis functions over the
            interval ``[-1, 1]``.

        """
        x, w, psi = self._x_ts_nodes, self._w_ts_nodes, self._psi_ts_nodes
        return self._quad_psi(fun, x, w, psi)

    @staticmethod
    def _gauss_legendre_rule(order: int) -> Tuple[Array, Array]:
        """Points and weights of the Gauss-Legendre quadrature.

        Returns
        -------
        x : array
            Quadrature points.
        w : array
            Quadrature weights.

        """
        x, w = roots_legendre(order)
        x = jnp.asarray(x)
        w = jnp.asarray(w)
        return x, w

    @staticmethod
    def _tanh_sinh_rule(order: int) -> Tuple[Array, Array]:
        """Points and weights of the tanh-sinh quadrature.

        Returns
        -------
        x : array
            Quadrature points.
        w : array
            Quadrature weights.

        """
        one_minus_eps = jnp.array(1.0) - 10 * jnp.finfo(jnp.array(1.0).dtype).eps
        tanh_inverse = lambda x: 0.5 * jnp.log((1 + x) / (1 - x))
        sinh_inverse = lambda x: jnp.log(x + jnp.sqrt(x**2 + 1))
        t_max = sinh_inverse(2. / jnp.pi * tanh_inverse(one_minus_eps))
        t = jnp.linspace(-t_max, t_max, num=order)

        x = jnp.tanh(0.5 * jnp.pi * jnp.sinh(t))
        w = 0.5 * jnp.pi * jnp.cosh(t) / jnp.cosh(0.5 * jnp.pi * jnp.sinh(t))**2
        w = w * jnp.diff(t)[0]
        w = 2 * w / jnp.sum(w)
        return x, w

    @classmethod
    def Legendre(cls, n_nodes: int, order: int = 101, order_nodes: int = 21) -> Self:
        """Legendre basis functions.

        Parameters
        ----------
        n_nodes : int
            Number of nodes. Corresponds to the order of the basis
            functions.
        order : int
            Order of the numerical quadratures over any interval
            ``[a, b]``.
        order_nodes : int
            Order of the numerical quadratures over the full interval
            ``[-1, 1]`` that use the node coordinates `xi` as stop points.

        Returns
        -------
        basis
            Instance of the `Basis` class with nodes located at
            Gauss-Legendre roots.

        """
        xi, w = roots_legendre(n_nodes)
        xi = jnp.array(xi)
        w = jnp.array(w)
        return cls(xi, order=order, order_nodes=order_nodes, w=w)
